program graupel
implicit none

! Ray tracing at ellipsoidal particle + Diffraction at spherical aperture
! with same geometrical cross section as spheroid for each orientation.
! Gives full phase matrix.

! Special thanks to Jouni Peltoniemi, whose rotation routine for the 
! Stokes-matrix has been adopted.

! Variable angular bin size for ray tracing AND diffraction 
! sideproduct: diffraction at exact angels: ifn + '.diff_discrete'
! NOTE : *.diff.discrete can differ from .diff if angular resolution is to small

! Reference: 
!!$@Article{MackeMishchenko96,
!!$  author = 	 "Macke, A. and Mishchenko, M. I.",
!!$  title = 	 "Applicability of regular particle shapes in
!!$		  light scattering calculations for atmospheric ice particles",
!!$  journal =	 AO,
!!$  year =	 1996,
!!$  volume =	 35,
!!$  pages =	 "4291-4296"
!!$}

! Input file (logical name 'ifn') must contain:
!   la, rbi, ibi : wavelength (\mu m), real & imag. part of refractive index
!   a_e, b_e, c_e : radius and length of circular cylinder
!   n_ray : Number of incident rays per crystal orientation
!   n_orient : Number of crystal orientations
!   n_r_max : max # of internal reflections 
!   rtddim :  # of angular bins - 1
!   diff_seperate = 1 if diffraction NOT added to ray tracing part, 0 else
!   no_diff = 1 if diffraction is omitted, 0 else

! Output files:
! ifn + '.status' : runtime status each 10000 rays, optical scalars like
!                   cross sections, single scattering albedo and 
!                   asymmetry parameter.
! ifn + '.save'   : saves intermediate results (binary!). Read with 'get.save.f'
! ifn + '.angles' : scattering angles == center angles of angular bins
! 
! ifn + '.1dr'    : phase function excluding diffraction
! ifn + '.diff', ifn + 'diff_discrete'   : diffraction pattern 
! ifn + '.p(r)11'    : total phase function 
! ifn + '.p(r)12'    : degree of linear polarization
!        .p(r)22/p(r)33/p(r)34/p(r)44 phase matrix elements
! 'r' instead of 'p' if diffraction is excluded from final output.

! Questions, comments, ... -> 
!!$  Andreas Macke, Leibniz Institute for Tropospheric Research


  integer(8), parameter ::   rtddim_max = 1800  ! max # of scatt. angle bins
  integer(8), parameter ::   angle_max = 10000! -> max # of angles -> Fresnel & Snell discretization 
  integer(8), parameter ::   scat_dim = 1000 ! # of scattering angle bins for internal scattering

  real(8) angle_scat(0:scat_dim), pf(0:scat_dim) ! scatt. angles, phase function
  real(8) angle_u(0:scat_dim), angle_o(0:scat_dim) ! scatt. angle bin boundaries
  real(8) arg1(0:scat_dim), pf_int_sum(0:scat_dim) ! solid angles, pf*arg
  real(8) d_pf_int_sum(0:scat_dim) ! delta(pf*arg)

  real(8) rtd(0:rtddim_max,1:4,1:4) ! scattered Stokes vectors 
  real(8) pij(0:rtddim_max,1:4,1:4) ! phase matrix
  real(8) diff_total(0:rtddim_max) ! diffraction function
  integer(8) bin_sum(0:rtddim_max) ! counts # of diffraction calc. per ang. bin
  real(8) ra(1:4,1:4,0:angle_max), ta(1:4,1:4,0:angle_max)
  real(8) rb(1:4,1:4,0:angle_max), tb(1:4,1:4,0:angle_max)  ! Fresnel coeff.
  real(8) rot_stokes(1:4,1:4), trans_r(1:4,1:4), trans_t(1:4,1:4)
! transf. matrices for rotation, reflection and transmission of Stokes v. 
  real(8) in_stokes(1:4,1:4), refl_stokes(1:4,1:4)
  real(8) new_stokes(1:4,1:4), new_in_stokes(1:4,1:4)
  real(8) trans_stokes(1:4,1:4), out_stokes(1:4,1:4) ! stokes
  real(8) rotate_in_stokes(1:4,1:4), rotate_out_stokes(1:4, 1:4)
  real(8) wfa(0:angle_max), wfb(0:angle_max)                       !s. PINIT 
  real(8) cos_theta_t_a(0:angle_max), cos_theta_t_b(0:angle_max)    !s. PINIT 
  real(8) x_hit(1:2), y_hit(1:2), z_hit(1:2)  
  real(8) xs_diff(1:2,1:3)
  real(8) x_i(1:3), x_i_z(1:3)
  real(8) sin_rot, cos_rot

  real(8) direction
  real(8) distortion ! 
  real(8) cos_theta, sin_theta, cos_phi, sin_phi
  real(8) w1, w2 ! help variable -> random azimuth 
  real(8) lost_tr ! lost due to finite # of total reflection 
  real(8) energy_violation ! :. in reflection/refraction ? 
  real(8) pi,twopi, pitwo                                       ! pi, 2*pi, pi/2 
  real(8) rad, deg                                               ! 180/pi, pi/180 
  real(8) d_the, d_phi, d_the_scat ! angular sampling resolution 
  real(8) energy_in, energy_out, energy_abs, abs_scat ! total incoming + scattered energies 
  real(8) resol_fak                             ! angular resolution for sampling 
  real(8) w_0, w_0_aerosol, co_w_0_aerosol ! single scattering albedo 
  real(8) g_total, g_rt, g_diff, g_diff_discrete
  real(8) grenzw, cosgrenzw             ! angle for total reflection , cos( :. ) 
  real(8) rayconst1                                 ! 2/pi*angle_max: for binning 
  real(8) help1, help2, help3, help4, help5, help6, help7, help8, test1, arg_scalar
  real(8) wegl                                       ! path length inside crystal 
  real(8) myfunc_arg  
  real(8) area_true ! geomtrical cross section 
  real(8) the_scat ! zenith scattering angle 
  real(8) absfak, mabs 
  real(8) ev11, ev12, ev13 ! e-vector perp 
  real(8) ev21, ev22, ev23 !           par
  real(8) ev31, ev32, ev33 ! prop. vector 
  real(8) f, fn1, fn2, fn3 ! surface normal
  real(8) rbi, ibi, la, nf ! refrative index, wavelength, normalization constant
  real(8) r, root
  real(8) t, the, phi
  real(8) vk11 , vk12 , vk13 ! e_perp...
  real(8) vk21 , vk22 , vk23 ! e_par...
  real(8) vk31 , vk32 , vk33 ! and prop for initial ray
  real(8) vk41 , vk42 , vk43 ! just in case
  real(8) vk71 , vk72 , vk73 ! e_perp...
  real(8) vk81 , vk82 , vk83 ! e_par...
  real(8) vk91 , vk92 , vk93 ! and prop for reflection
  real(8) vk101, vk102, vk103 ! e_perp...
  real(8) vk111, vk112, vk113 ! e_par...
  real(8) vk121, vk122, vk123 ! and prop for refraction
  real(8) hv1, hv2, hv3 ! perp to scattering plane
  real(8) w, wd, wf, wfz, wvz ! angles
  real(8) a_e, b_e, c_e, sqr_a_e, sqr_b_e, sqr_c_e ! ellipsoid parameters
  real(8) x_s, y_s, z_s, x_s_old, y_s_old, z_s_old ! points on ellipsoid
  real(8) x_s_start, y_s_start, z_s_start ! dito
  real(8) theta_i, phi_i, angle
  real(8) p, p1
  real(8) a_root, b_root, c_root, t1, t2, sqrtroot
  real(8) x_aufpunkt, y_aufpunkt, z_aufpunkt
  real(8) h11, h12, h13, h21, h22, h23
  real(8) kh1, kh2, kh3 ! 
  real(8) k_wave, x, k_wave2 ! 2 pi /la , size parameter
  real(8) a ! ###@##
  real(8) sum, sum1, sum2
  real(8) x_orient, area_orient, radius_orient
  real(8) fourp2, rad2resol_fak
  real(8) rand
  real(8) mean_fpl, fpl, real_path_ray, real_path ! various pathlenghts
  real(8) k_x, k_y, k_z, theta_0_true, costheta_0, sintheta_0 ! 
  real(8) sinphi_0, cosphi_0 ! jumpandscatter
  real pintegral
  real jump_wegl, total_wegl
  real(8) start_time_cpu, end_time_cpu

  integer(8) dummy, n_miss, n_try !
  integer(8) n_miss_orient, n_try_orient
  integer(8) angle_i                               ! address for incident angle 
  integer(8) angle_g                    !     "       angle of total reflection 
  integer(8) counter ! in tilt
  integer(8) rtddim ! # of scattering angle bins
  integer(8) n_r_max ! max # of internal reflections
  integer(8) h, i, j, k, l, nr, nt, nr_sum, nt_sum, ntz, i0
  integer(8) ntt, ntm, seed, n_the_scat
  integer(8) n_ray, i_ray, n_orient, i_orient
  integer(8) pure_str ! get rid of trailing blanks  
  integer(8) start_time, end_time, irtc
  integer(8) no_diff, diff_seperate  
  integer(8) jump_count ! counts internal scattering events
  integer(8) theta_0 ! another integer angle 
  integer(8) scatt_angle_max ! max. # of internal scattering angles
  character(100) ifn, ofn, afn
 
  call SDATIN ! Data input: wavelength, ellipsoid, orientation :.  
  call PINIT  ! Init. of constants, limit. values, and sum. variables

  do i_orient = 1, n_orient ! orientation loop
    n_try_orient = 0
    n_miss_orient = 0
    call wave_front ! initializes wave front r + (h1,h2)
  
    do i_ray = 0, n_ray  ! ray loop 
      call init_ray ! first incoming ray: direction, e-field
      call PRTA ! External reflection + refraction into the ellipsoid

! Each ray inside the ellipsoid is traced as long as the 
!  remaining energy 'in_stokes(1,1)' is smaller then 1.0e-10 or the number of
!  internal reflections 'nr' exceeds n_r_max 

      nr = 0 ! counts internal reflections
      nt = 0 ! counts internal total reflections
      Do while (in_stokes(1,1) .gt. 1.0e-10) !  .and. nr .lt. n_r_max)
        nr = nr + 1
        call PINTERSECT ! next point of intersec.
        real_path_ray = real_path_ray + wegl*in_stokes(1,1)
        call random_number(rand) ! jump first
        if (rand .gt. 0) then 
          fpl = -mean_fpl*log(rand)
        else 
          fpl = 1e+10 ! ### 
        endif
        jump_wegl = 0. ! cummulative path length from internal jumps 
        do while (fpl .lt. wegl)
          jump_count = jump_count + 1
          call jumpandscatter ! x_s, y_s, z_s, ev3
          jump_wegl = jump_wegl + fpl
          call PINTERSECT ! -> x_s, ..., wegl
          call random_number(rand)
          if (rand .gt. 0) then 
            fpl = -mean_fpl*log(rand)
          else
            fpl = 1e+10 ! ### 
          endif
        end do
        total_wegl = jump_wegl + wegl
! = 0 + (wegl from first call to PINTERSECT) or jump_wegl + (wegl from last call to PINTERSECT) 
        absfak = exp(mabs * total_wegl)     ! Absorption 
        call PRTB ! Outgoing refr. and internal refl. 
        x_s_old = x_s
        y_s_old = y_s
        z_s_old = z_s
      end do ! internal reflections
!      if (nr .ge. n_r_max) lost_tr = lost_tr + in_stokes(1,1)
      nr_sum = nr_sum + nr ! -> mean # of internal reflections
      nt_sum = nt_sum + nt ! -> mean # of internal total reflections
      if (ntz .gt. 100000) then
        call PZS ! store the momentary results 
        ntz = 0
      end if
    end do                    ! ray loop 
    call diff_orient
  end do ! orientation loop
  call PZS                  ! just in case PZS wasn't called before (w_0 = 0 then)
  call PRTNORM ! normalize phase function, phase matrix, g
  call PZS ! store the final results 

  contains

  real(8) FUNCTION BESSJ1(X)
  REAL*8 Y,P1,P2,P3,P4,P5,Q1,Q2,Q3,Q4,Q5,R1,R2,R3,R4,R5,R6,S1,S2,S3,S4,S5,S6
  real*8 x, xx, ax, z, ans1, ans2
    DATA R1,R2,R3,R4,R5,R6/72362614232.D0,-7895059235.D0,242396853.1D0, &
    -2972611.439D0,15704.48260D0,-30.16036606D0/, &
    S1,S2,S3,S4,S5,S6/144725228442.D0,2300535178.D0, &
    18583304.74D0,99447.43394D0,376.9991397D0,1.D0/
    DATA P1,P2,P3,P4,P5/1.D0,.183105D-2,-.3516396496D-4,.2457520174D-5, &
    -.240337019D-6/, Q1,Q2,Q3,Q4,Q5/.04687499995D0,-.2002690873D-3, &
    .8449199096D-5,-.88228987D-6,.105787412D-6/
    IF(ABS(X).LT.8.)THEN
      Y=X**2
      ans1 = X*(R1+Y*(R2+Y*(R3+Y*(R4+Y*(R5+Y*R6)))))
      ans2 = (S1+Y*(S2+Y*(S3+Y*(S4+Y*(S5+Y*S6)))))
      bessj1 = ans1/ans2
    ELSE
      AX=ABS(X)
      Z=8./AX
      Y=Z**2
      XX=AX-2.356194491
      ans1 = (P1+Y*(P2+Y*(P3+Y*(P4+Y*P5))))
      ans2 = (Q1+Y*(Q2+Y*(Q3+Y*(Q4+Y*Q5))))
      BESSJ1=SQRT(.636619772/AX)*(COS(XX)*ans1-Z*SIN(XX)*ans2)*SIGN(1d0,X)
    ENDIF
  end function bessj1

  real(8) function pcirc(x, tb)
  real(8) x, y, tb
    help1 = x*(1 + cos(tb))
    arg_scalar = x*sin(tb)
    help2 = bessj1(arg_scalar)/arg_scalar ! numerical recipes
    pcirc = help1*help1*help2*help2 
  end function pcirc 

  subroutine wave_front
    call random_number(rand)
    cos_theta = 1. - 2.*rand ! cos(theta)
    sin_theta = sqrt(1. - cos_theta*cos_theta) ! sin(theta)
    test1 = 2
    do while (test1 .gt. 1) ! effective search for sin/cos(phi)...
      call random_number(rand)
      w1 = 1. - 2.*rand ! IBM
      call random_number(rand)
      w2 = 1. - 2.*rand ! ... if phi is random ! IBM
      test1 = w1*w1 + w2*w2 
    end do
    help1 = sqrt(test1)
    cos_phi = w1/help1
    sin_phi = w2/help1
    fn1 = cos_phi*sin_theta ! normal vector to wave front
    fn2 = sin_phi*sin_theta
    fn3 = cos_theta
    x_aufpunkt = p*fn1 ! random start point on sphere 
    y_aufpunkt = p*fn2 ! surrounding the ellipsoid
    z_aufpunkt = p*fn3

    if (abs(fn3) .lt. 1.0) then
      h11 = fn2
      h12 = -fn1
      h13 = 0.0
      nf = sqrt(h11*h11 + h12*h12)
      h11 = h11/nf*p
      h12 = h12/nf*p
      h21 = fn3*fn1
      h22 = fn3*fn2
      h23 = -fn2*fn2 - fn1*fn1
      nf = sqrt(h21*h21 + h22*h22 + h23*h23)
      h21 = h21/nf*p
      h22 = h22/nf*p
      h23 = h23/nf*p
    else
      h11 = 1
      h12 = 0
      h13 = 0
      h21 = 0
      h22 = 1
      h23 = 0
    end if

    vk31 = -fn1 ! incoming direction for this wave front realization
    vk32 = -fn2
    vk33 = -fn3
    vk11 = vk42 * vk33 - vk43 * vk32 ! E_perp 
    vk12 = vk43 * vk31 - vk41 * vk33
    vk13 = vk41 * vk32 - vk42 * vk31
    nf = sqrt(vk11*vk11 + vk12*vk12 + vk13*vk13)
    vk11 = vk11 / nf 
    vk12 = vk12 / nf 
    vk13 = vk13 / nf

  end subroutine wave_front

  subroutine init_ray
    root = -1.

    ev31 = vk31 ! get original direction...
    ev32 = vk32
    ev33 = vk33
    ev11 = vk11 ! ... and polarization
    ev12 = vk12
    ev13 = vk13
    
    do while (root .lt. 0.0) ! seek for a hit...
      n_try = n_try + 1
      n_try_orient = n_try_orient + 1
      call random_number(rand)
      t1 = 1.0 - 2.0*rand
      call random_number(rand)
      t2 = 1.0 - 2.0*rand
      x_s_start = x_aufpunkt + t1*h11 + t2*h21 ! random point on wave_front-
      y_s_start = y_aufpunkt + t1*h12 + t2*h22 ! square
      z_s_start = z_aufpunkt + t1*h13 + t2*h23
      a_root = ev31*ev31/sqr_a_e + ev32*ev32/sqr_b_e + ev33*ev33/sqr_c_e
      b_root = 2*(x_s_start*ev31/sqr_a_e + &
                  y_s_start*ev32/sqr_b_e + &
                  z_s_start*ev33/sqr_c_e)
      c_root = x_s_start*x_s_start/sqr_a_e + &
               y_s_start*y_s_start/sqr_b_e + &
               z_s_start*z_s_start/sqr_c_e - 1.0
      root = b_root*b_root - 4*a_root*c_root
      if (root .lt. 0) then
        n_miss = n_miss + 1
        n_miss_orient = n_miss_orient + 1
      end if
    end do ! while
    sqrtroot = sqrt(root)
    help1 = 2*a_root
    t2 = (-b_root + sqrtroot) / help1
    t1 = (-b_root - sqrtroot) / help1
    t = min(t1, t2)
    x_s = x_s_start + t*ev31
    y_s = y_s_start + t*ev32
    z_s = z_s_start + t*ev33
    x_s_old = x_s 
    y_s_old = y_s 
    z_s_old = z_s 

    ntt = ntt + 1 
    ntz = ntz + 1 ! Temp. number of Rays 
  end subroutine init_ray      


  subroutine PINIT ! initializes this and that 
  real(8) seed
  real(8) nr, ni, nr1, ni1, nr2, ni2, k1, k2, krel, ref1, ref2, ref3
  real(8) arg, fac1, theta_i, theta_t, ref6, nrel
  real(8) r11, r12, r33, r34, t11, t12, t33, t34

    call system_clock(start_time)
    call cpu_time(start_time_cpu)

    n_miss = 0
    n_try = 0
    nr_sum = 0
    nt_sum = 0
    bin_sum = 0

    call random_seed

    pi = 4*atan(1.0) 
    twopi = 2.*pi 
    pitwo = pi/2.  
    deg = 180./pi 
    rad = pi/180.
    resol_fak = dble(rtddim)/180. ! 1/resol_fak = angular resolution in degree 
    rad2resol_fak = rad/resol_fak
    k_wave = 2.*pi/la
    k_wave2 = k_wave*k_wave
    x = k_wave*a ! size parameter
    rtd = 0.
    diff_total = 0.

    vk41 = 1. / sqrt(3.)
    vk42 = 1. / sqrt(3.)
    vk43 = 1. / sqrt(3.)

    sqr_a_e = a_e*a_e 
    sqr_b_e = b_e*b_e 
    sqr_c_e = c_e*c_e 

    energy_in = 0. 
    energy_out = 0. 
    energy_abs = 0.
    abs_scat = 0.
    ntt = 0

    angle_scat(0) = 0.0
    angle_scat(scatt_angle_max + 1) = 180.0

    i = 0
    angle_o(i) = 0.0
    do while (angle_o(i) .lt. 180.0)
      i = i + 1
      angle_u(i) = angle_o(i - 1)
      angle_o(i) = 2*angle_scat(i) - angle_u(i)
      arg1(i) = cos(angle_u(i)*rad) - cos(angle_o(i)*rad) ! d_Omega
    end do

    sum = 0.0
    pf_int_sum(0) = 0.0
    do i = 1, scatt_angle_max
      d_pf_int_sum(i) = pf(i)*arg1(i)
      sum = sum + d_pf_int_sum(i)
      pf_int_sum(i) = sum
    end do 
    
    pintegral = pf_int_sum(scatt_angle_max)

    do i = 1, scatt_angle_max
      pf_int_sum(i) = pf_int_sum(i)/pintegral
    end do
    do i = 1, scatt_angle_max
      d_pf_int_sum(i) = pf_int_sum(i) - pf_int_sum(i-1)
    end do

    p1 = max(a_e,b_e) 
    p = max(p1, c_e)
    fourp2 = 4.*p*p
    mabs = -4*Pi/la*ibi ! absorption coefficient
    myfunc_arg = 1/rbi*1/Sqrt( 1 - ibi*ibi/rbi/rbi)
    if (abs(myfunc_arg) .gt. 1.0) myfunc_arg = sign(1d0, myfunc_arg)
    grenzw = asin(myfunc_arg) ! angle of total reflection and 
    angle_g = int(grenzw*2/pi*angle_max) ! corresp. bin 
    cosgrenzw = cos(grenzw)
    rayconst1 = 2/pi*angle_max

    rot_stokes = 0
    ra = 0
    ta = 0
    rb = 0
    tb = 0
    rot_stokes(1,1) = 1 
    rot_stokes(4,4) = 1

    nr1 = 1.0
    ni1 = 0.0
    nr2 = rbi
    ni2 = ibi

    k1 = ni1/nr1
    k2 = ni2/nr2
    krel = (k2 - k1)/(1 + k1*k2)
    nrel = nr2/nr1*(1 + k1*k2)/(1 + k1*k1)
    ref1 = nrel*nrel
    ref2 = krel*krel
    ref3 = (1 + ref2)*(1 + ref2)
    ref6 = ref1*ref3/(1 + krel*k2)/(1 + krel*k2)
    fac1 = pi/2./angle_max

    do i = 0, angle_max - 1
      theta_i = (i + 0.5)*fac1
      if (theta_i .le. pi/2) then 
        call refraction(ref1, ref2, ref3, ref6, theta_i, theta_t, krel, k2)
call fresnel(nr1, ni1, nr2, ni2, theta_i, r11, r12, r33, r34, t11, t12, t33, t34)
        help5 = 0.5 
        help6 = 0.5*cos(theta_t)*nr2/nr1/cos(theta_i)

        ra(1,1,i) = help5*r11
        ra(1,2,i) = help5*r12 
        ra(3,3,i) = help5*r33 
        ra(3,4,i) = help5*r34
        ta(1,1,i) = help6*t11
        ta(1,2,i) = help6*t12
        ta(3,3,i) = help6*t33
        ta(3,4,i) = help6*t34
        ra(2,2,i) = ra(1,1,i)
        ra(2,1,i) = ra(1,2,i)
        ra(4,4,i) = ra(3,3,i)
        ra(4,3,i) =-ra(3,4,i)
        ta(2,2,i) = ta(1,1,i)
        ta(2,1,i) = ta(1,2,i)
        ta(4,4,i) = ta(3,3,i)
        ta(4,3,i) =-ta(3,4,i)

        w = dcos(theta_i)
        help1 = sqrt(1 - w*w)
        if (help1 .eq. 0) then 
          wfa(i) = 0 ! ###
        else
          wfa(i) = dsin(theta_t)/help1
        endif
        cos_theta_t_a(i) = dcos(theta_t)

      endif
    enddo

    nr1 = rbi
    ni1 = ibi
    nr2 = 1.0
    ni2 = 0.0
    k1 = ni1/nr1
    k2 = ni2/nr2
    krel = (k2 - k1)/(1 + k1*k2)
    nrel = nr2/nr1*(1 + k1*k2)/(1 + k1*k1)
    ref1 = nrel*nrel
    ref2 = krel*krel
    ref3 = (1 + ref2)*(1 + ref2)
    ref6 = ref1*ref3/(1 + krel*k2)/(1 + krel*k2)

    do i = 0, angle_max - 1
      theta_i = (i + 0.5)*fac1
      if (theta_i .le. grenzw) then ! Refrac. angle & Fresnel-mat.: Ref. & Trans  
        call refraction(ref1, ref2, ref3, ref6, theta_i, theta_t, krel, k2)

call fresnel(nr1, ni1, nr2, ni2, theta_i, r11, r12, r33, r34, t11, t12, t33, t34)

        help5 = 0.5 
        help6 = 0.5*cos(theta_t)*nr2/nr1/cos(theta_i)
 
        rb(1,1,i) = help5*r11
        rb(1,2,i) = help5*r12
        rb(3,3,i) = help5*r33
        rb(3,4,i) = help5*r34
        tb(1,1,i) = help6*t11
        tb(1,2,i) = help6*t12
        tb(3,3,i) = help6*t33
        tb(3,4,i) = help6*t34

        rb(2,2,i) = rb(1,1,i)
        rb(2,1,i) = rb(1,2,i)
        rb(4,4,i) = rb(3,3,i)
        rb(4,3,i) =-rb(3,4,i)
        tb(2,2,i) = tb(1,1,i)
        tb(2,1,i) = tb(1,2,i)
        tb(4,4,i) = tb(3,3,i)
        tb(4,3,i) =-tb(3,4,i)

        w = dcos(theta_i)
        help1 = sqrt(1 - w*w)
        if (help1 .eq. 0) then 
          wfb(i) = 0
        else
          wfb(i) = dsin(theta_t)/help1
        endif
        cos_theta_t_b(i) = dcos(theta_t)

      else ! Fresnel-matr. for total internal reflection

call fresnel(nr1, ni1, nr2, ni2, theta_i, r11, r12, r33, r34, t11, t12, t33, t34)
        help5 = 0.5

        rb(1,1,i) = help5*r11
        rb(1,2,i) = help5*r12
        rb(3,3,i) = help5*r33
        rb(3,4,i) = help5*r34
        rb(2,2,i) = rb(1,1,i)
        rb(2,1,i) = rb(1,2,i)
        rb(4,4,i) = rb(3,3,i)
        rb(4,3,i) =-rb(3,4,i)
      end if
    end do
  end subroutine pinit

  subroutine refraction(ref1, ref2, ref3, ref6, theta_i, theta_t, krel, k2)

  real(8) ref1, ref2, ref3, ref6, theta_i, theta_t, krel, k2
  real(8) sintiq, ref4, ref5, q4, q2, g, test1, test2
  real(8) ref7, rnstar

!  Calculates the angle of refraction for transmission from 
!  (nr1, ni1) -> (nr2,ni2)

    sintiq = sin(theta_i)*sin(theta_i)
    ref4 = 1 - (1 - ref2)/ref1/ref3*sintiq
    ref5 = 2*krel/ref1/ref3*sintiq
    q4 = ref4*ref4 + ref5*ref5
    q2 = sqrt(q4)
    g = asin(ref5/q2)/2
    test1 = acos(ref4/q2)/2
    test2 = atan(ref5/ref4)/2
    g = test1
    ref7 = (cos(g) - k2*sin(g))*(cos(g) - k2*sin(g))
    rnstar = sqrt(sintiq + ref6*q2*ref7)        
    theta_t = asin(sin(theta_i)/rnstar)

  end subroutine refraction

  subroutine fresnel(nr1, ni1, nr2, ni2, theta_i, r11, r12, r33, r34, t11, t12, t33, t34)

!   input: complex refractive index of medium 1: cn1 = nr1 + i ni1
!                         "                   2: cn2 = nr2 + i ni2
!          angle of incidence in rad. : theta_i
!
!   output: fresnel coefficients for transforming field amplitudes
!           r11, r12, r33, r34 : reflection
!           t11, t12, t33, t34 : refraction
  real(8) nr1, ni1, nr2, ni2, theta_i, sinti
  real(8) r11, r12, r33, r34, t11, t12, t33, t34
  complex(8) ci, cn1, cn2, csintiq, costi, chelp1, cts, crs, ctp, crp

     ci = dcmplx(0,-1) 
     cn1 = dcmplx(nr1, ni1)
     cn2 = dcmplx(nr2, ni2)
     sinti = dsin(theta_i)
     csintiq = dcmplx(sinti*sinti)
     costi = dcmplx(cos(theta_i))
     chelp1 = cdsqrt(cn2*cn2 - cn1*cn1*csintiq)
       
     cts = 2*cn1*costi/(cn1*costi + chelp1)
     crs = (cn1*costi - chelp1)/(cn1*costi + chelp1)
     ctp = 2*cn1*cn2*costi/(cn2*cn2*costi + cn1*chelp1)
     crp = (cn2*cn2*costi - cn1*chelp1)/(cn2*cn2*costi + cn1*chelp1)

     r11 = Real(crp*Conjg(crp) + crs*Conjg(crs))
     r12 = Real(crp*Conjg(crp) - crs*Conjg(crs))
     r33 = 2.*Real(crp*Conjg(crs))
     r34 = 2.*Aimag(crp*Conjg(crs))
     t11 = Real(ctp*Conjg(ctp) + cts*Conjg(cts))
     t12 = Real(ctp*Conjg(ctp) - cts*Conjg(cts))
     t33 = 2.*Real(ctp*Conjg(cts))
     t34 = 2.*Aimag(ctp*Conjg(cts))

  end subroutine fresnel


  subroutine SDATIN ! Reads in the Input data 
    read(5,*) ifn
    h = 1
    do while (ifn(h:h) .ne. ' ')
      h = h + 1
    end do
    pure_str = h - 1 ! store for all other concatenations
    open(unit = 10, file = ifn(1:pure_str), status = 'old', form = 'formatted')
    read(10, fmt = '(A100)') afn ! phase function 
    open(unit = 11, file = afn, status = 'old', form = 'formatted')
    i = 0 
    do while (1 .lt. 2)
      i = i + 1
      read(11,*, end = 99) angle_scat(i), pf(i)
    end do
 99    continue ! herrje
    scatt_angle_max = i - 1
    close(11)
    read(10,*) w_0_aerosol
    co_w_0_aerosol = 1.d0 - w_0_aerosol
    read(10,*) mean_fpl
    read(10,*) la, rbi, ibi ! wavelength + corresp. refractive index 
    read(10,*) a_e, b_e, c_e ! Ellipsoid semi-axis. 
    read(10,*) distortion ! surface distortion in [0,1]
! NOTE: For spheroids: a_e = b_e 
    read(10,*) n_ray ! # of rays
    read(10,*) n_orient
    read(10,*) n_r_max ! max. number of internal reflections
    read(10,*) rtddim ! sampling resolution
    read(10,*) diff_seperate
    read(10,*) no_diff
    close(10) 
  end subroutine SDATIN
 
  subroutine PINTERSECT 
! :.computes the intersection (x,y,z) of a given ellipsoid 
!     x^2/a^2 + y^2/b^2 + z^2/c^2 = 1
!  and a given ray
!     (x,y,z) = (x_s, y_s, z_s) + t*(ev31, ev32, ev33)
    a_root = ev31*ev31/sqr_a_e + ev32*ev32/sqr_b_e + ev33*ev33/sqr_c_e
    b_root = 2.d0*(x_s_old*ev31/sqr_a_e + y_s_old*ev32/sqr_b_e + z_s_old*ev33/sqr_c_e)
    c_root = x_s_old*x_s_old/sqr_a_e + y_s_old*y_s_old/sqr_b_e + z_s_old*z_s_old/sqr_c_e - 1.d0
! c_root = ... always = 0 when (x_s, y_s, z_s) already on the ellipsoid
! -> use modified PINTERSECT for non-scattering cases ###########
    root = b_root*b_root - 4*a_root*c_root
    sqrtroot = sqrt(root)
    help1 = 2.d0*a_root
    t2 = (-b_root + sqrtroot) / help1
    t1 = (-b_root - sqrtroot) / help1
    t = max(t1,t2)

    x_s = x_s_old + t*ev31
    y_s = y_s_old + t*ev32
    z_s = z_s_old + t*ev33
test1 = x_s*x_s/a_e/a_e + y_s*y_s/b_e/b_e + z_s*z_s/c_e/c_e - 1.0 ! ###
!if (test1 .gt. 1.e-9) print *, test1
    wegl = t
  end subroutine pintersect

  subroutine PFN ! Surface normal at the point of intersection 
    fn1 = x_s / sqr_a_e 
    fn2 = y_s / sqr_b_e
    fn3 = z_s / sqr_c_e
    nf = sqrt( fn1*fn1 + fn2*fn2 + fn3*fn3 )
    fn1 = fn1/nf
    fn2 = fn2/nf
    fn3 = fn3/nf
  end subroutine PFN

  subroutine rotation_matrix(en1, en2, en3, eo1, eo2, eo3, ko1, ko2, ko3)
! Adopted from Jouni Peltoniemi
  real(8) eo1, eo2, eo3 ! rotation_matrix: perp. to reference plane
  real(8) en1, en2, en3 !        "       : perp. e-field 
  real(8) ko1, ko2, ko3 !        '       : direction

    myfunc_arg = en1*eo1 + en2*eo2 + en3*eo3 ! perp_new * perp_old
    ev21 = ko2*en3 - ko3*en2 
    ev22 = ko3*en1 - ko1*en3
    ev23 = ko1*en2 - ko2*en1
    test1 = myfunc_arg*(eo1*ev21 + eo2*ev22 + eo3*ev23) ! perp_new * par_old 
    help1 = myfunc_arg*myfunc_arg
    cos_rot = min(1.0d+0, 2.*help1 - 1.)
    if (test1 .gt. 0.) then 
      sin_rot = sqrt(1. - cos_rot*cos_rot)
    else
      sin_rot = -sqrt(1. - cos_rot*cos_rot)
    end if
    rot_stokes(2,2) = cos_rot
    rot_stokes(2,3) =-sin_rot
    rot_stokes(3,2) = sin_rot
    rot_stokes(3,3) = cos_rot
  end subroutine rotation_matrix

  subroutine PRTA ! Ray Tracing for external rays 
    call PFN ! surface normal vector at the point of intersection 
    call tilt
    if (w .lt. -1.) w = -1 ! numerical reasons 

    help1 = abs(w)
    angle_i = int(acos(help1)*rayconst1) ! angel bin 

    vk91 = ev31 + 2.0 * help1 * fn1 ! direction vector for reflection 
    vk92 = ev32 + 2.0 * help1 * fn2
    vk93 = ev33 + 2.0 * help1 * fn3   
    nf = SQRT(vk91 * vk91 + vk92 * vk92 + vk93 * vk93)
    vk91 = vk91 / nf 
    vk92 = vk92 / nf 
    vk93 = vk93 / nf

    vk71 = fn2*ev33 - fn3*ev32 ! vector perp. to plane of incidence
    vk72 = fn3*ev31 - fn1*ev33
    vk73 = fn1*ev32 - fn2*ev31
    nf = SQRT(vk71*vk71 + vk72*vk72 + vk73*vk73)
    vk71 = vk71/nf 
    vk72 = vk72/nf 
    vk73 = vk73/nf

    wvz = w / abs(w)
    wf = wfa(angle_i)
    help2 = cos_theta_t_a(angle_i)
      !  ---- direction of refraction ----  
    vk121 = wf * (ev31 + help1 * fn1) + wvz * help2 * fn1
    vk122 = wf * (ev32 + help1 * fn2) + wvz * help2 * fn2
    vk123 = wf * (ev33 + help1 * fn3) + wvz * help2 * fn3
    nf = SQRT(vk121 * vk121 + vk122 * vk122 + vk123 * vk123)
    vk121 = vk121 / nf 
    vk122 = vk122 / nf 
    vk123 = vk123 / nf

    call rotation_matrix(vk71, vk72, vk73, ev11, ev12, ev13, ev31, ev32, ev33)
    new_in_stokes = rot_stokes ! ### only if initial stokes matrix is ID-matrix

    do i = 1, 4
      do j = 1, 4
        trans_r(i,j) = ra(i,j,angle_i)
        trans_t(i,j) = ta(i,j,angle_i)
      end do
    end do

    refl_stokes = matmul(trans_r, new_in_stokes)
    trans_stokes = matmul(trans_t, new_in_stokes)

    energy_in = energy_in + 1.
    test1 = 1. - refl_stokes(1,1) - trans_stokes(1,1) ! ### 
    energy_violation = energy_violation + abs(test1) ! ### 

! SAMPLING: 
    help1 = vk91*vk31 + vk92*vk32 + vk93*vk33
    the_scat = acos(help1)*rtddim/pi
    n_the_scat = nint(the_scat) ! scattering zenith angle 
    energy_out = energy_out + refl_stokes(1,1)

    out_stokes = matmul(refl_stokes, transpose(new_in_stokes))

    do i0 = 1, 4  ! stokes vectors 
      do i = 1,4  ! stokes vect. components 
        rtd(n_the_scat, i0, i) = rtd(n_the_scat, i0, i) +  out_stokes(i0,i)
      end do
    end do

    do i0 = 1, 4  ! Hand-over for PRTB 
      do i = 1, 4
        in_stokes(i0,i) = trans_stokes(i0,i)
      end do
    end do
    ev11 = vk71 ! perp.
    ev12 = vk72
    ev13 = vk73

    ev31 = vk121 ! prop.
    ev32 = vk122
    ev33 = vk123
    
  end subroutine prta

  subroutine PRTB ! Ray Tracing for internal rays 

    call PFN ! surface normal at the point of intersection 
    call tilt
    if (w .gt. 1) w = 1 ! numerical reasons 
    angle_i = int(acos(w)*rayconst1)!  angel bin 

    help1 = abs(w)
    vk91 = ev31 - 2.0 * help1 * fn1 ! direction vector of 
    vk92 = ev32 - 2.0 * help1 * fn2 ! reflected ray 
    vk93 = ev33 - 2.0 * help1 * fn3  
    nf = SQRT(vk91 * vk91 + vk92 * vk92 + vk93 * vk93)
    vk91 = vk91 / nf 
    vk92 = vk92 / nf 
    vk93 = vk93 / nf

    vk71 = fn2*ev33 - fn3*ev32 ! vector perp. to plane of incidence
    vk72 = fn3*ev31 - fn1*ev33
    vk73 = fn1*ev32 - fn2*ev31
    nf = SQRT(vk71*vk71 + vk72*vk72 + vk73*vk73)
    vk71 = vk71/nf 
    vk72 = vk72/nf 
    vk73 = vk73/nf

    if (angle_i .lt. angle_g) then ! if no total reflection then :. 

      wvz = w / ABS(w) ! ### evtl. ist if :. schneller 
      wf = wfb(angle_i)
      help1 = cos_theta_t_b(angle_i)
        !  ---- direction of propagation ----  
      vk121 = wf * (ev31 - w * fn1) + wvz * help1 * fn1
      vk122 = wf * (ev32 - w * fn2) + wvz * help1 * fn2
      vk123 = wf * (ev33 - w * fn3) + wvz * help1 * fn3
      nf = SQRT(vk121 * vk121 + vk122 * vk122 + vk123 * vk123)
      vk121 = vk121 / nf 
      vk122 = vk122 / nf 
      vk123 = vk123 / nf

      call rotation_matrix(vk71, vk72, vk73, ev11, ev12, ev13, ev31, ev32, ev33)

      new_in_stokes = matmul(rot_stokes, in_stokes)

      do i = 1, 4
        do j = 1, 4
          trans_r(i,j) = rb(i,j,angle_i)
          trans_t(i,j) = tb(i,j,angle_i)
        end do
      end do

      refl_stokes = matmul(trans_r, new_in_stokes)
      trans_stokes = matmul(trans_t, new_in_stokes)

      test1 = in_stokes(1,1)  - refl_stokes(1,1) - trans_stokes(1,1)! ### 
      energy_violation = energy_violation + abs(test1) ! ### 

      energy_abs = energy_abs + (1 - absfak)*in_stokes(1,1)

!SAMPLING:                 
      hv1 = vk32*vk123 - vk33*vk122 ! 
      hv2 = vk33*vk121 - vk31*vk123 ! hv perp. to scattering plane 
      hv3 = vk31*vk122 - vk32*vk121 ! 
      nf = sqrt(hv1*hv1 + hv2*hv2 + hv3*hv3)
      hv1 = -hv1/nf ! k x k0
      hv2 = -hv2/nf
      hv3 = -hv3/nf ! may be hvex -> -hv

      call rotation_matrix(vk11, vk12, vk13, hv1, hv2, hv3, vk31, vk32, vk33)
! Rotate initial Stokes matrix into scattering plane

      rotate_in_stokes = rot_stokes

      call rotation_matrix(hv1, hv2, hv3, vk71, vk72, vk73, vk121, vk122, vk123)
! 2nd rotation

      rotate_out_stokes = rot_stokes

      out_stokes = matmul(rotate_out_stokes, matmul(trans_stokes, rotate_in_stokes))
      help1 = vk121*vk31 + vk122*vk32 + vk123*vk33
      the_scat = acos(help1)*rtddim/pi
      n_the_scat = nint(the_scat) ! scattering zenith angle 

      do i0 = 1, 4  ! stokes vectors 
        do i = 1,4  ! stokes vect. components 
          rtd(n_the_scat, i0, i) = rtd(n_the_scat, i0, i) &
          +  out_stokes(i0,i)*absfak
        end do
      end do
      energy_out = energy_out + out_stokes(1,1)*absfak

! end sampling

    else                                       ! Totalreflexion 

      nt = nt + 1 ! # of total reflections per ray
      call rotation_matrix(vk71, vk72, vk73, ev11, ev12, ev13, ev31, ev32, ev33)

      new_in_stokes = matmul(rot_stokes, in_stokes)

      do i = 1, 4
        do j = 1, 4
          trans_r(i,j) = rb(i,j,angle_i)
        end do
      end do

      refl_stokes = matmul(trans_r, new_in_stokes)

      test1 = in_stokes(1,1)  - refl_stokes(1,1) ! ### 
      energy_violation = energy_violation + abs(test1) ! ### 

      energy_abs = energy_abs + (1 - absfak)*in_stokes(1,1)

    end if

    do i0 = 1, 4  ! Hand-over for next call to PRTB 
      do i = 1,4 
        in_stokes(i0,i) = refl_stokes(i0,i)*absfak
      end do
    end do

    ev11 = vk71 ! hand over direction vectors
    ev12 = vk72
    ev13 = vk73       
    ev31 = vk91
    ev32 = vk92
    ev33 = vk93

  end subroutine PRTB

  subroutine PZS            ! store intermediate status 
  real(8) ext_cs, ext_eff, abs_cs, abs_eff, scat_cs, scat_eff, r_eq, x_eq
  integer(8) i , j, k
  character(26) date

   call date_and_time(date)
    call system_clock(end_time)
    call cpu_time(end_time_cpu)

    ext_eff = 2. ! extinction efficiency = 2, what else 
    ext_cs = ext_eff*area_true ! " cross section 
    scat_eff = 1 + (energy_in - energy_abs - abs_scat)/energy_in ! ### -> S. 226
! Diffraction eff. (==1) + raytracing eff. 
    scat_cs  = area_true*scat_eff
    abs_eff = ext_eff - scat_eff ! absorption efficiency 
    abs_cs = abs_eff*area_true ! absorption cross section 
    
    w_0 = scat_eff/ext_eff ! single scattering albedo 
    r_eq = sqrt(area_true/pi)
    x_eq = 2.*pi*r_eq/la
    
    open(12, file = ifn(1:pure_str)//'.status', form = 'formatted')
!    write(12,*) ' User ', log_name, 'on ', maschine_name, ' in ', cwd_name
    write(12,*) ' Inputfile: ', ifn
    write(12,*) ' Date & time: ', date
    write(12,*) ' Elapsed time: ', int((end_time - start_time)/1.e9), ' sec.' ! time
    write(12,*) ' Elapsed CPU time: ', int(end_time_cpu - start_time_cpu), ' sec.' ! time
    write(12,*) ' '
    write(12,*) i_orient - 1, ' out of ', n_orient, 'orientations '
!    write(12,*) ntt, ' hits out of ', n_orient*n_ray, ' incoming rays'
    write(12,*) dble(jump_count)/dble(ntt), ' internal scattering events per ray'
    write(12,*) ' '
    write(12,*) 'Incoming energy     : ', energy_in
    write(12,*) 'Scattered energy    : ', energy_out
    write(12,*) 'Real absorbed energy: ', energy_abs
help1 = energy_in - energy_out - energy_abs
    write(12,*) 'Missing energy      : ',help1,' = ',help1/energy_in*100, '%'
    write(12,*) 'FreSnell error      : ', energy_violation ! ### 
    write(12,*) 'finite # int. refls : ', lost_tr
    write(12,*) 'Internal (total) reflections per ray:', dble(nr_sum)/dble(ntt), '(', dble(nt_sum)/dble(ntt),')'
    write(12,*) 'radius of surface equiv. sphere: ',r_eq
    write(12,*) 'size parameter: ', x_eq
    write(12,*) ' '
    write(12,*) '                  extinction, scattering, absorption : '
    write(12,*) 'cross sections: ',ext_cs,' ',scat_cs,' ',abs_cs
    write(12,*) 'efficiencies  : ',ext_eff,' ',scat_eff,' ',abs_eff
    write(12,*) ' '
    write(12,*) 'w_0    : ', w_0
    write(12,*) 'g_rt   : ', g_rt
    write(12,*) 'g_diff (g_diff_discrete) : ', g_diff, g_diff_discrete
    write(12,*) 'g_total: ', g_total       
    write(12,*) ' '
    write(12,*) 'Ellipsoids projected area: ', area_true
    help1 = 4.*p*p*dble((n_try - n_miss))/dble(n_try)
    write(12,*) 'Scanning projected area  : ', help1
    close(12)

  end subroutine PZS

  subroutine diff_orient ! diffraction at circular aperture with same cross sectional area
  integer(8) i

    area_orient = fourp2*dble((n_try_orient - n_miss_orient))/dble(n_try_orient) 
    radius_orient = sqrt(area_orient/pi) ! radius
    x_orient = k_wave*radius_orient ! size parameter

    diff_total(0) = diff_total(0) + k_wave2*radius_orient*radius_orient ! exact forward diffraction

    do i = 1, rtddim - 1
      angle = dble(i)*rad2resol_fak
      diff_total(i) = diff_total(i) + pcirc(x_orient,angle)
    end do

    diff_total(rtddim) = diff_total(rtddim - 1) ! avoid "0" 

  end subroutine diff_orient


  subroutine ellipsoid_surface
! numerical integration of ellipsoid surface 

integer(8), parameter :: itmax = 500 ! theta integration
integer(8), parameter :: ipmax = 500 ! phi integration
integer(8), parameter :: newton_cotes_max = 2 ! Simpson

real(8) a, b, c, t, p, dt, dp, a2, b2, c2, a2b2
real(8) st, st2, st4, ct, ct2, sp, sp2, cp, cp2
real(8) p_integral, t_integral
real(8) p_integrand(0:ipmax), t_integrand(0:itmax)
real(8) newton_cotes_fak(0:newton_cotes_max)
real(8) newton_cotes_fak_p, newton_cotes_fak_t
integer(8) it, ip, is, newton_cotes_pn

  newton_cotes_pn = 6
  newton_cotes_fak(0) = 1.
  newton_cotes_fak(1) = 4.
  newton_cotes_fak(2) = 1.

  newton_cotes_fak_p = 90.*rad/ipmax/newton_cotes_pn
  newton_cotes_fak_t = 90.*rad/itmax/newton_cotes_pn

  a2b2 = a_e*a_e*b_e*b_e
  a2 = a_e*a_e
  b2 = b_e*b_e
  c2 = c_e*c_e

  do it = 0, itmax
    t = 90./itmax*rad*it
    st = sin(t)
    st2 = st*st
    st4 = st2*st2
    ct = cos(t)
    ct2 = ct*ct
    do ip = 0, ipmax
      p = 90./ipmax*rad*ip
      sp = sin(p)
      sp2 = sp*sp
      cp = cos(p)
      cp2 = cp*cp
      p_integrand(ip) = sqrt(a2b2*st2*ct2 + c2*st4*(a2*sp2 + b2*cp2))
    end do
    p_integral = 0.
    do ip = 0, ipmax - newton_cotes_max
      do is = 0, newton_cotes_max ! Trapez, Simpson, ...
        p_integral = p_integral + p_integrand(ip + is)*newton_cotes_fak(is) 
      end do
    end do
    p_integral = p_integral*newton_cotes_fak_p
    t_integrand(it) = p_integral
  end do
  t_integral= 0.
  do it = 0, itmax - newton_cotes_max
    do is = 0, newton_cotes_max ! Trapez, Simpson, ...
      t_integral = t_integral + t_integrand(it + is)*newton_cotes_fak(is) 
    end do
  end do
  t_integral = t_integral*newton_cotes_fak_t

  area_true = t_integral*2.

end subroutine ellipsoid_surface

  subroutine PRTNORM ! final normalization and output 
  real(8) d_o1, ptotal, rt11, diff, x1, x2, y1, y2, x, y, a, b, exc
  integer(8) i, j, k

    if (a_e .eq. b_e .and. a_e .gt. c_e) then ! oblate spheroid
      a = a_e
      b = c_e
      exc = 1./a*sqrt(a*a - b*b)
      area_true = 0.5*pi*a*a*(1. + (1. - exc*exc)/2./exc*log((1. + exc)/(1. - exc)))
    end if 
    if (a_e .eq. b_e .and. a_e .lt. c_e) then ! prolate spheroid
      a = c_e
      b = a_e
      exc = 1./a*sqrt(a*a - b*b)
      area_true = 0.5*pi*b*(b + a*asin(exc)/exc)
    end if
    if (a_e .eq. b_e .and. b_e .eq. c_e) then ! sphere
      area_true = pi*a_e*a_e
    end if
    if (a_e .ne. b_e .and. b_e .ne. c_e) then ! real ellipsoid
      call ellipsoid_surface
    end if

    help1 = rad/resol_fak ! delta theta 
    help2 = rad*0.25d0/resol_fak
    help3 = rad*(180.d0 - 0.25d0/resol_fak) 
    help5 = 2.d0*w_0 - 1.d0
    help6 = 2.d0*w_0
    help7 = help5/help6

    if (no_diff .eq. 0) then 

      call diff_discrete ! normalize and write "discrete" diffraction phase function

! Interpolate diffraction at center angle of bin 1 ...
      y1 = diff_total(0)
      y2 = diff_total(1)
      x1 = 0.
      x2 = 1./resol_fak
      x = help2*deg
      y = (y1 - y2)/(x1 - x2)*(x - x1) + y1
      diff_total(0) = y ! ... and replace exact forward diff. with this
! Interpolate diffraction at center angle of bin rtddim ...
      y1 = diff_total(rtddim - 1)
      y2 = diff_total(rtddim)
      x1 = (rtddim - 1)*help1*deg
      x2 = 180
      x = help3*deg
      y = (y1 - y2)/(x1 - x2)*(x - x1) + y1
      diff_total(rtddim) = y ! ... and replace exact backward diff. with this
! all additional diffraction values can be regarded as bin means. 

      sum1 = 0 ! normalize diffraction ...
      do i = 0, rtddim 
        d_o1 = 2*sin(i*help1)*sin(help1/2) ! solid angle interval 
        if (i .eq. 0) d_o1 = 2*sin(help2)*sin(help2)
        if (i .eq. rtddim) d_o1 = 2*sin(help3)*sin(help2)
        sum1 = sum1 + diff_total(i)*d_o1 ! dE/dOmega * dOmega = dE
      end do
      do i = 0, rtddim
        diff_total(i) = diff_total(i)/sum1*2.
      end do

      open(63, file = ifn(1:pure_str)//'.diff')  ! Diffraction pattern 
      g_diff = 0.
      do i = 0, rtddim 
        angle = i*help1*deg
        d_o1 = 2*sin(i*help1)*sin(help1/2) ! solid angle interval 
        if (i .eq. 0) then 
          angle = help2*deg
          d_o1 = 2*sin(help2)*sin(help2)
        end if
        if (i .eq. rtddim) then
          angle = help3*deg
          d_o1 = 2*sin(help3)*sin(help2)
        end if
        write(63, 1004) angle, diff_total(i)
        help4 = cos(angle*rad)*d_o1
        g_diff = g_diff + diff_total(i)*help4
      end do
      g_diff = g_diff/2.
      close(63)
    end if ! (no_diff = 0) if

    sum2 = 0. ! ... and ray tracing
    do i = 0, rtddim 
      sum2 = sum2 + rtd(i,1,1) ! == dE
    end do

    open(61, file = ifn(1:pure_str)//'.1dr') ! ray tracing phase function: rt11 
    do i = 0, rtddim ! divide dE by solid angle interval -> phase function
      d_o1 = 2*sin(i*help1)*sin(help1/2) ! solid angle interval 
      angle = i*help1*deg
      if (i .eq. 0) then 
        d_o1 = 2*sin(help2)*sin(help2)
        angle = help2*deg
      end if
      if (i .eq. rtddim) then
        d_o1 = 2*sin(help3)*sin(help2)
        angle = help3*deg
      end if
      help4 = cos(angle*rad)*d_o1

      do j = 1, 4
        do k = 1, 4
          rtd(i,j,k) = rtd(i,j,k)/d_o1/sum2*2.
       end do
      end do
      write(61, 1004) angle, rtd(i,1,1)
      g_rt = g_rt + rtd(i,1,1)*help4
    end do
    close(61)
    g_rt = g_rt/2. 


 1003 format(f8.4,1x,f7.4)
 1004 format(f8.4,1x,e9.3)

    if (diff_seperate .eq. 0 .and. no_diff .eq. 0) then 

      open(12, file = ifn(1:pure_str)//'.p12')  
!     open(21, file = ifn(1:pure_str)//'.p21') ! only for testing purposes
      open(22, file = ifn(1:pure_str)//'.p22')
      open(33, file = ifn(1:pure_str)//'.p33')
      open(34, file = ifn(1:pure_str)//'.p34')
!     open(43, file = ifn(1:pure_str)//'.p43') ! only for testing purposes
      open(44, file = ifn(1:pure_str)//'.p44')
!     open(25, file = ifn(1:pure_str)//'.pij.bin') ! only for testing purposes
      open(64, file = ifn(1:pure_str)//'.p11') ! total phase function 

      g_total = 0. ! asymmetry parameter 
      do i = 0, rtddim 
        angle = i*help1*deg
        d_o1 = 2*sin(i*help1)*sin(help1/2) ! solid angle interval 
        if (i .eq. 0) then 
          angle = help2*deg
          d_o1 = 2*sin(help2)*sin(help2)
        end if
        if (i .eq. rtddim) then
          angle = help3*deg
          d_o1 = 2*sin(help3)*sin(help2)
        end if

        rt11 = rtd(i,1,1)
        diff = diff_total(i)
        ptotal = (diff + help5*rt11)/help6

        do j = 1, 4 ! add diffraction to diagonal elements
          do k = 1, 4
            if (j .eq. k) then 
              pij(i,j,k) = (diff + help5*rtd(i,j,k))/help6
            else 
              pij(i,j,k) = rtd(i,j,k)*help7
            end if
          end do
        end do
        write(64, 1004) angle, ptotal


!       write(25) angle, ((pij(i,j,k)/pij(i,1,1), j = 1,4), k = 1,4)
        write(12, 1003) angle, pij(i,1,2)/pij(i,1,1)
!       write(21, 1003) angle, pij(i,2,1)/pij(i,1,1)
        write(22, 1003) angle, pij(i,2,2)/pij(i,1,1)
        write(33, 1003) angle, pij(i,3,3)/pij(i,1,1)
        write(34, 1003) angle, pij(i,3,4)/pij(i,1,1)
!       write(43, 1003) angle, pij(i,4,3)/pij(i,1,1)
        write(44, 1003) angle, pij(i,4,4)/pij(i,1,1)

        help4 = cos(angle*rad)*d_o1
        g_total = g_total + ptotal*help4
      end do
!     close(25)
      close(12)
!     close(21)
      close(22)
      close(33)
      close(34)
!     close(43)
      close(44)
      close(64)
      g_total = g_total/2.0

    else ! pure ray tracing

      open(12, file = ifn(1:pure_str)//'.r12')  
!     open(21, file = ifn(1:pure_str)//'.r21') ! only for testing purposes
      open(22, file = ifn(1:pure_str)//'.r22')
      open(33, file = ifn(1:pure_str)//'.r33')
      open(34, file = ifn(1:pure_str)//'.r34')
!     open(43, file = ifn(1:pure_str)//'.r43') ! only for testing purposes
      open(44, file = ifn(1:pure_str)//'.r44')
!     open(25, file = ifn(1:pure_str)//'.rij.bin') ! only for testing purposes
      do i = 0, rtddim
        angle = i*help1*deg
        if (i .eq. 0) angle = help2*deg
        if (i .eq. rtddim) angle = help3*deg
!       write(25) angle, ((rtd(i,j,k)/rtd(i,1,1), j = 1,4), k = 1,4)
        write(12, 1003) angle, rtd(i,1,2)/rtd(i,1,1)
!       write(21, 1003) angle, rtd(i,2,1)/rtd(i,1,1)
        write(22, 1003) angle, rtd(i,2,2)/rtd(i,1,1)
        write(33, 1003) angle, rtd(i,3,3)/rtd(i,1,1)
        write(34, 1003) angle, rtd(i,3,4)/rtd(i,1,1)
!       write(43, 1003) angle, rtd(i,4,3)/rtd(i,1,1)
        write(44, 1003) angle, rtd(i,4,4)/rtd(i,1,1)      
      end do

!     close(25)
      close(12)
!     close(21)
      close(22)
      close(33)
      close(34)
!     close(43)
      close(44)

    end if

  end subroutine PRTNORM  

  subroutine diff_discrete
  integer(8) i, j, iangle
  real(8) simpson_fak(0:2), simpson, sintheta, costheta

! The following would be a good integration of the diffraction function. 
! Unfortunately, it does not fit to the 'binned' ray tracing results.
! We do it anyway and save it under inf + '.diff.discrete' 
! The corresponding asymmetry parameter -> g_diff_discrete

    help1 = rad/resol_fak ! delta theta 
    help2 = rad*0.25/resol_fak
    help3 = rad*(180. - 0.25/resol_fak) 

    simpson = 2.*pi/dble(rtddim)/6.
    simpson_fak(0) = 1.
    simpson_fak(1) = 4.
    simpson_fak(2) = 1.

    sum = 0.
    do i = 0, rtddim - 2, 2! test diffraction phase function normalization
      do j = 0, 2
        iangle = i + j
        angle = dble(iangle)*rad2resol_fak
        sintheta = sin(angle)
        sum = sum + diff_total(iangle)*sintheta*simpson_fak(j)
      end do
    end do
    sum = sum*simpson ! seems to be = 2pi !!!

    do i = 0, rtddim ! renormalize diffraction phase function
      diff_total(i) = diff_total(i)/sum*2.
    end do

    g_diff_discrete = 0
    do i = 0, rtddim - 2, 2 ! calculate asymmetry parameter
      do j = 0, 2
        iangle = i + j
        angle = dble(iangle)*rad2resol_fak
        sintheta = sin(angle)
        costheta = cos(angle)
        g_diff_discrete = g_diff_discrete + diff_total(iangle)*costheta*sintheta*simpson_fak(j)
      end do
    end do
    g_diff_discrete = g_diff_discrete*simpson/2.

    open(10, file = ifn(1:pure_str)//'.diff.discrete')
    do i = 0, rtddim 
      angle = i*help1*deg
      write(10,1004) angle, diff_total(i)
    end do
    close(10)
 1004 format(f8.4,1x,e9.3)

  end subroutine diff_discrete

  subroutine tilt
    counter = 0 ! random distortion of the rays propagation vector
    w = 1.0 
    do while (w .ge. 0 .and. counter .lt. 10)  ! equivalent to displacemet
      counter = counter + 1  ! of crystal coordinates
      call random_number(rand)
      kh1 = ev31 + distortion*(2*rand-1)
      call random_number(rand)
      kh2 = ev32 + distortion*(2*rand-1)
      call random_number(rand)
      kh3 = ev33 + distortion*(2*rand-1)
      test1 = sqrt(kh1*kh1 + kh2*kh2 + kh3*kh3)
      kh1 = kh1/test1 
      kh2 = kh2/test1 
      kh3 = kh3/test1
      w = kh1*fn1 + kh2*fn2 + kh3*fn3 !-dcos(theta_i
    end do

    if (counter .lt. 10) then
      ev31 = kh1 
      ev32 = kh2 
      ev33 = kh3
    else ! w = negative cosine of angle of incidence 
      w = ev31*fn1 + ev32*fn2 + ev33*fn3
    end if

  end subroutine tilt

  subroutine jumpandscatter

    x_s_old = x_s_old + fpl*ev31 ! new photon location 
    y_s_old = y_s_old + fpl*ev32 
    z_s_old = z_s_old + fpl*ev33 

    theta_0 = 1 
    call random_number(rand)
    do while (pf_int_sum(theta_0) .lt. rand) 
      theta_0 = theta_0 + 1
    end do

    t = (rand - pf_int_sum(theta_0 - 1))/d_pf_int_sum(theta_0) 
! (r - S(n-1))/(S(n) - S(n-1)) in [0,1] -> interpolation
    theta_0_true = angle_u(theta_0) + t*(angle_o(theta_0) - angle_u(theta_0))

    help1 = theta_0_true*rad
    costheta_0 = cos(help1)
    sintheta_0 = sin(help1)

! check if above process reproduces the input scattering function
!    theta_0 = idnint(theta_0_true)
!    theta_hauf(theta_0) = theta_hauf(theta_0) + 1.0

    test1 = 2
    do while (test1 .gt. 1) ! effective search for sin/cos(phi)...
      call random_number(rand)
      w1 = 1 - 2*rand
      call random_number(rand)
      w2 = 1 - 2*rand ! ... if phi is random ! IBM
      test1 = w1*w1 + w2*w2 
    end do
    help1 = sqrt(test1)
    cosphi_0 = w1/help1
    sinphi_0 = w2/help1

    k_x = ev31
    k_y = ev32
    k_z = ev33

    test1 = 1 - k_z*k_z ! sin^2(theta_old)
    if (test1 .gt. 0) then ! old photon direction not exactly along z-axis
      help1 = sqrt((1 - costheta_0*costheta_0)/test1)
      ev31 = k_x*costheta_0 - (k_y*sinphi_0 + k_x*k_z*cosphi_0)*help1
      ev32 = k_y*costheta_0 + (k_x*sinphi_0 - k_y*k_z*cosphi_0)*help1
      ev33 = k_z*costheta_0 + test1*cosphi_0*help1
    else
      ev31 = sintheta_0*cosphi_0
      ev32 = sintheta_0*sinphi_0
      ev33 = k_z*costheta_0
    end if    

    help1 = in_stokes(1,1)*co_w_0_aerosol ! intensity loss (better ENERGY loss..)
    abs_scat = abs_scat + help1
    in_stokes(1,1) = in_stokes(1,1) - help1

  end subroutine jumpandscatter  

end program graupel







